# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""
Tests for lava.commands.
"""

import os
import tempfile

from mock import (
    MagicMock,
    patch
)

from lava.commands import (
    init,
    submit,
)
from lava.config import Config
from lava.helper.tests.helper_test import HelperTest
from lava.tool.errors import CommandError


class InitCommandTests(HelperTest):

    def setUp(self):
        super(InitCommandTests, self).setUp()
        self.config_file = self.tmp("init_command_tests")
        self.config = Config()
        self.config.config_file = self.config_file

    def tearDown(self):
        super(InitCommandTests, self).tearDown()
        if os.path.isfile(self.config_file):
            os.unlink(self.config_file)

    def test_register_arguments(self):
        self.args.DIR = os.path.join(tempfile.gettempdir(), "a_fake_dir")
        init_command = init(self.parser, self.args)
        init_command.register_arguments(self.parser)

        # Make sure we do not forget about this test.
        self.assertEqual(2, len(self.parser.method_calls))

        _, args, _ = self.parser.method_calls[0]
        self.assertIn("--non-interactive", args)

        _, args, _ = self.parser.method_calls[1]
        self.assertIn("DIR", args)

    @patch("lava.commands.edit_file", create=True)
    def test_command_invoke_0(self, mocked_edit_file):
        # Invoke the init command passing a path to a file. Should raise an
        # exception.
        self.args.DIR = self.temp_file.name
        init_command = init(self.parser, self.args)
        self.assertRaises(CommandError, init_command.invoke)

    def test_command_invoke_2(self):
        # Invoke the init command passing a path where the user cannot write.
        try:
            self.args.DIR = "/root/a_temp_dir"
            init_command = init(self.parser, self.args)
            self.assertRaises(CommandError, init_command.invoke)
        finally:
            if os.path.exists(self.args.DIR):
                os.removedirs(self.args.DIR)

    def test_update_data(self):
        # Make sure the template is updated accordingly with the provided data.
        self.args.DIR = self.temp_file.name

        init_command = init(self.parser, self.args)
        init_command.config.get = MagicMock()
        init_command.config.save = MagicMock()
        init_command.config.get.side_effect = ["a_job.json"]

        expected = {
            "jobfile": "a_job.json",
        }

        obtained = init_command._update_data()
        self.assertEqual(expected, obtained)


class SubmitCommandTests(HelperTest):

    def setUp(self):
        super(SubmitCommandTests, self).setUp()
        self.config_file = self.tmp("submit_command_tests")
        self.config = Config()
        self.config.config_file = self.config_file
        self.config.save = MagicMock()

    def tearDown(self):
        super(SubmitCommandTests, self).tearDown()
        if os.path.isfile(self.config_file):
            os.unlink(self.config_file)

    def test_register_arguments(self):
        self.args.JOB = os.path.join(tempfile.gettempdir(), "a_fake_file")
        submit_command = submit(self.parser, self.args)
        submit_command.register_arguments(self.parser)

        # Make sure we do not forget about this test.
        self.assertEqual(2, len(self.parser.method_calls))

        _, args, _ = self.parser.method_calls[0]
        self.assertIn("--non-interactive", args)

        _, args, _ = self.parser.method_calls[1]
        self.assertIn("JOB", args)



# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""lava_tool.utils tests."""

import os
import shutil
import subprocess
import sys
import tempfile

from unittest import TestCase
from mock import (
    MagicMock,
    call,
    patch,
)

from lava.tool.errors import CommandError
from lava_tool.utils import (
    can_edit_file,
    create_dir,
    edit_file,
    execute,
    has_command,
    retrieve_file,
    verify_and_create_url,
    verify_file_extension,
)


class UtilTests(TestCase):

    def setUp(self):
        self.original_stdout = sys.stdout
        sys.stdout = open("/dev/null", "w")
        self.original_stderr = sys.stderr
        sys.stderr = open("/dev/null", "w")
        self.original_stdin = sys.stdin
        self.temp_file = tempfile.NamedTemporaryFile(delete=False)

    def tearDown(self):
        sys.stdin = self.original_stdin
        sys.stdout = self.original_stdout
        sys.stderr = self.original_stderr
        os.unlink(self.temp_file.name)

    @patch("lava_tool.utils.subprocess.check_call")
    def test_has_command_0(self, mocked_check_call):
        # Make sure we raise an exception when the subprocess is called.
        mocked_check_call.side_effect = subprocess.CalledProcessError(0, "")
        self.assertFalse(has_command(""))

    @patch("lava_tool.utils.subprocess.check_call")
    def test_has_command_1(self, mocked_check_call):
        # Check that a "command" exists. The call to subprocess is mocked.
        mocked_check_call.return_value = 0
        self.assertTrue(has_command(""))

    def test_verify_file_extension_with_extension(self):
        extension = ".test"
        supported = [extension[1:]]
        try:
            temp_file = tempfile.NamedTemporaryFile(suffix=extension,
                                                    delete=False)
            obtained = verify_file_extension(
                temp_file.name, extension[1:], supported)
            self.assertEquals(temp_file.name, obtained)
        finally:
            if os.path.isfile(temp_file.name):
                os.unlink(temp_file.name)

    def test_verify_file_extension_without_extension(self):
        extension = "json"
        supported = [extension]
        expected = "/tmp/a_fake.json"
        obtained = verify_file_extension("/tmp/a_fake", extension, supported)
        self.assertEquals(expected, obtained)

    def test_verify_file_extension_with_unsupported_extension(self):
        extension = "json"
        supported = [extension]
        expected = "/tmp/a_fake.json"
        obtained = verify_file_extension(
            "/tmp/a_fake.extension", extension, supported)
        self.assertEquals(expected, obtained)

    @patch("os.listdir")
    def test_retrieve_job_file_0(self, mocked_os_listdir):
        # Make sure that exception is raised if we go through all the elements
        # returned by os.listdir().
        mocked_os_listdir.return_value = ["a_file"]
        self.assertRaises(CommandError, retrieve_file,
                          "a_path", ["ext"])

    @patch("os.listdir")
    def test_retrieve_job_file_1(self, mocked_os_listdir):
        # Pass some files and directories to retrieve_file(), and make
        # sure a file with .json suffix is returned.
        # Pass also a hidden file.
        try:
            json_file = tempfile.NamedTemporaryFile(suffix=".json")
            json_file_name = os.path.basename(json_file.name)

            file_name_no_suffix = tempfile.NamedTemporaryFile(delete=False)
            file_name_with_suffix = tempfile.NamedTemporaryFile(
                suffix=".bork", delete=False)

            temp_dir_name = "submit_command_test_tmp_dir"
            temp_dir_path = os.path.join(tempfile.gettempdir(), temp_dir_name)
            os.makedirs(temp_dir_path)

            hidden_file = tempfile.NamedTemporaryFile(
                prefix=".tmp", delete=False)

            mocked_os_listdir.return_value = [
                temp_dir_name, file_name_no_suffix.name,
                file_name_with_suffix.name, json_file_name, hidden_file.name]

            obtained = retrieve_file(tempfile.gettempdir(), ["json"])
            self.assertEqual(json_file.name, obtained)
        finally:
            os.removedirs(temp_dir_path)
            os.unlink(file_name_no_suffix.name)
            os.unlink(file_name_with_suffix.name)
            os.unlink(hidden_file.name)

    def test_retrieve_job_file_2(self):
        # Pass a file with the valid extension.
        temp_file = tempfile.NamedTemporaryFile(suffix=".json")
        obtained = retrieve_file(temp_file.name, ["json"])
        self.assertEquals(temp_file.name, obtained)

    def test_retrieve_job_file_3(self):
        # Pass a file with a non-valid extension.
        temp_file = tempfile.NamedTemporaryFile(suffix=".bork")
        self.assertRaises(
            CommandError, retrieve_file, temp_file.name, ["json"])

    @patch("os.listdir")
    def test_retrieve_job_file_4(self, mocked_os_listdir):
        # Pass hidden and wrong files and make sure exception is thrown.
        a_hidden_file = ".a_hidden.json"
        b_hidden_file = ".b_hidden.json"
        c_wrong_file = "a_wrong_file.bork"

        mocked_os_listdir.return_value = [a_hidden_file, b_hidden_file, c_wrong_file]
        self.assertRaises(
            CommandError, retrieve_file, tempfile.gettempdir(), ["json"])

    @patch("lava_tool.utils.subprocess")
    def test_execute_0(self, mocked_subprocess):
        mocked_subprocess.check_call = MagicMock()
        execute("foo")
        self.assertEqual(mocked_subprocess.check_call.call_args_list,
                         [call(["foo"])])
        self.assertTrue(mocked_subprocess.check_call.called)

    @patch("lava_tool.utils.subprocess.check_call")
    def test_execute_1(self, mocked_check_call):
        mocked_check_call.side_effect = subprocess.CalledProcessError(1, "foo")
        self.assertRaises(CommandError, execute, ["foo"])

    @patch("lava_tool.utils.subprocess")
    @patch("lava_tool.utils.has_command", return_value=False)
    @patch("lava_tool.utils.os.environ.get", return_value=None)
    @patch("lava_tool.utils.sys.exit")
    def test_edit_file_0(self, mocked_sys_exit, mocked_env_get,
                         mocked_has_command, mocked_subprocess):
        edit_file(self.temp_file.name)
        self.assertTrue(mocked_sys_exit.called)

    @patch("lava_tool.utils.subprocess")
    @patch("lava_tool.utils.has_command", side_effect=[True, False])
    @patch("lava_tool.utils.os.environ.get", return_value=None)
    def test_edit_file_1(self, mocked_env_get, mocked_has_command,
                         mocked_subprocess):
        mocked_subprocess.Popen = MagicMock()
        edit_file(self.temp_file.name)
        expected = [call(["sensible-editor", self.temp_file.name])]
        self.assertEqual(expected, mocked_subprocess.Popen.call_args_list)

    @patch("lava_tool.utils.subprocess")
    @patch("lava_tool.utils.has_command", side_effect=[False, True])
    @patch("lava_tool.utils.os.environ.get", return_value=None)
    def test_edit_file_2(self, mocked_env_get, mocked_has_command,
                         mocked_subprocess):
        mocked_subprocess.Popen = MagicMock()
        edit_file(self.temp_file.name)
        expected = [call(["xdg-open", self.temp_file.name])]
        self.assertEqual(expected, mocked_subprocess.Popen.call_args_list)

    @patch("lava_tool.utils.subprocess")
    @patch("lava_tool.utils.has_command", return_value=False)
    @patch("lava_tool.utils.os.environ.get", return_value="vim")
    def test_edit_file_3(self, mocked_env_get, mocked_has_command,
                         mocked_subprocess):
        mocked_subprocess.Popen = MagicMock()
        edit_file(self.temp_file.name)
        expected = [call(["vim", self.temp_file.name])]
        self.assertEqual(expected, mocked_subprocess.Popen.call_args_list)

    @patch("lava_tool.utils.subprocess")
    @patch("lava_tool.utils.has_command", return_value=False)
    @patch("lava_tool.utils.os.environ.get", return_value="vim")
    def test_edit_file_4(self, mocked_env_get, mocked_has_command,
                         mocked_subprocess):
        mocked_subprocess.Popen = MagicMock()
        mocked_subprocess.Popen.side_effect = Exception()
        self.assertRaises(CommandError, edit_file, self.temp_file.name)

    def test_can_edit_file(self):
        # Tests the can_edit_file method of the config command.
        # This is to make sure the device config file is not erased when
        # checking if it is possible to open it.
        expected = ("hostname = a_fake_panda02\nconnection_command = \n"
                    "device_type = panda\n")

        with open(self.temp_file.name, "w") as f:
            f.write(expected)

        self.assertTrue(can_edit_file(self.temp_file.name))
        obtained = ""
        with open(self.temp_file.name) as f:
            obtained = f.read()

        self.assertEqual(expected, obtained)

    def test_verify_and_create_url_0(self):
        expected = "https://www.example.org/"
        obtained = verify_and_create_url("www.example.org")
        self.assertEquals(expected, obtained)

    def test_verify_and_create_url_1(self):
        expected = "http://www.example.org/"
        obtained = verify_and_create_url("http://www.example.org")
        self.assertEquals(expected, obtained)

    def test_verify_and_create_url_2(self):
        expected = "http://www.example.org/RPC/"
        obtained = verify_and_create_url("http://www.example.org/RPC")
        self.assertEquals(expected, obtained)

    def test_verify_and_create_url_3(self):
        expected = "https://www.example.org/RPC/"
        obtained = verify_and_create_url("www.example.org/RPC")
        self.assertEquals(expected, obtained)

    def test_create_dir_0(self):
        try:
            temp_dir = os.path.join(tempfile.gettempdir(), "a_dir")
            create_dir(temp_dir)
            self.assertTrue(os.path.isdir(temp_dir))
        finally:
            shutil.rmtree(temp_dir)

    def test_create_dir_1(self):
        try:
            temp_dir = os.path.join(tempfile.gettempdir(), "a_dir")
            create_dir(temp_dir, "subdir")
            self.assertTrue(os.path.isdir(os.path.join(temp_dir, "subdir")))
        finally:
            shutil.rmtree(temp_dir)

    def test_create_dir_2(self):
        temp_dir = os.path.join("/", "a_temp_dir")
        self.assertRaises(CommandError, create_dir, temp_dir)



# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""
lava.parameter unit tests.
"""

from mock import patch

from lava.helper.tests.helper_test import HelperTest
from lava.parameter import (
    ListParameter,
    Parameter,
    SingleChoiceParameter,
)
from lava_tool.utils import to_list


class GeneralParameterTest(HelperTest):
    """General class with setUp and tearDown methods for Parameter tests."""
    def setUp(self):
        super(GeneralParameterTest, self).setUp()
        # Patch class raw_input, start it, and stop it on tearDown.
        self.patcher1 = patch("lava.parameter.raw_input", create=True)
        self.mocked_raw_input = self.patcher1.start()

    def tearDown(self):
        super(GeneralParameterTest, self).tearDown()
        self.patcher1.stop()


class ParameterTest(GeneralParameterTest):
    """Tests for the Parameter class."""

    def setUp(self):
        super(ParameterTest, self).setUp()
        self.parameter1 = Parameter("foo", value="baz")

    def test_prompt_0(self):
        # Tests that when we have a value in the parameters and the user press
        # Enter, we get the old value back.
        self.mocked_raw_input.return_value = "\n"
        obtained = self.parameter1.prompt()
        self.assertEqual(self.parameter1.value, obtained)

    def test_prompt_1(self,):
        # Tests that with a value stored in the parameter, if and EOFError is
        # raised when getting user input, we get back the old value.
        self.mocked_raw_input.side_effect = EOFError()
        obtained = self.parameter1.prompt()
        self.assertEqual(self.parameter1.value, obtained)

    def test_to_list_0(self):
        value = "a_value"
        expected = [value]
        obtained = to_list(value)
        self.assertIsInstance(obtained, list)
        self.assertEquals(expected, obtained)

    def test_to_list_1(self):
        expected = ["a_value", "b_value"]
        obtained = to_list(expected)
        self.assertIsInstance(obtained, list)
        self.assertEquals(expected, obtained)


class ListParameterTest(GeneralParameterTest):

    """Tests for the specialized ListParameter class."""

    def setUp(self):
        super(ListParameterTest, self).setUp()
        self.list_parameter = ListParameter("list")

    def test_prompt_0(self):
        # Test that when pressing Enter, the prompt stops and the list is
        # returned.
        expected = []
        self.mocked_raw_input.return_value = "\n"
        obtained = self.list_parameter.prompt()
        self.assertEqual(expected, obtained)

    def test_prompt_1(self):
        # Tests that when passing 3 values, a list with those values
        # is returned
        expected = ["foo", "bar", "foobar"]
        self.mocked_raw_input.side_effect = expected + ["\n"]
        obtained = self.list_parameter.prompt()
        self.assertEqual(expected, obtained)

    def test_serialize_0(self):
        # Tests the serialize method of ListParameter passing a list.
        expected = "foo,bar,baz,1"
        to_serialize = ["foo", "bar", "baz", "", 1]

        obtained = self.list_parameter.serialize(to_serialize)
        self.assertEqual(expected, obtained)

    def test_serialize_1(self):
        # Tests the serialize method of ListParameter passing an int.
        expected = "1"
        to_serialize = 1

        obtained = self.list_parameter.serialize(to_serialize)
        self.assertEqual(expected, obtained)

    def test_deserialize_0(self):
        # Tests the deserialize method of ListParameter with a string
        # of values.
        expected = ["foo", "bar", "baz"]
        to_deserialize = "foo,bar,,baz,"
        obtained = self.list_parameter.deserialize(to_deserialize)
        self.assertEqual(expected, obtained)

    def test_deserialize_1(self):
        # Tests the deserialization method of ListParameter passing a list.
        expected = ["foo", 1, "", "bar"]
        obtained = self.list_parameter.deserialize(expected)
        self.assertEqual(expected, obtained)

    def test_set_value_0(self):
        # Pass a string to a ListParameter, expect a list.
        set_value = "foo"
        expected = [set_value]
        self.list_parameter.set(set_value)
        self.assertEquals(expected, self.list_parameter.value)

    def test_set_value_1(self):
        # Pass a list to a ListParameter, expect the same list.
        expected = ["foo", "bar"]
        self.list_parameter.set(expected)
        self.assertEquals(expected, self.list_parameter.value)

    def test_add_value_0(self):
        # Add a value to a ListParameter, expect a list back.
        add_value = "foo"
        expected = [add_value]
        self.list_parameter.add(add_value)
        self.assertEquals(expected, self.list_parameter.value)

    def test_add_value_1(self):
        # Add a list value to a ListParameter with already a value set, expect
        # a list with both values.
        # The ListParameter is initialized with a string.
        add_value = ["foo"]
        list_param = ListParameter("list", value="bar")
        expected = ["bar", "foo"]
        list_param.add(add_value)
        self.assertEquals(expected, list_param.value)

    def test_add_value_2(self):
        # Add a list value to a ListParameter with already a value set, expect
        # a list with both values.
        # The ListParameter is initialized with a list.
        add_value = ["foo"]
        list_param = ListParameter("list", value=["bar", "baz"])
        expected = ["bar", "baz", "foo"]
        list_param.add(add_value)
        self.assertEquals(expected, list_param.value)


class TestsSingleChoiceParameter(GeneralParameterTest):

    def setUp(self):
        super(TestsSingleChoiceParameter, self).setUp()
        self.choices = ["foo", "bar", "baz", "bam"]
        self.param_id = "single_choice"
        self.single_choice_param = SingleChoiceParameter(self.param_id,
                                                         self.choices)

    def test_with_old_value(self):
        # There is an old value for a single choice parameter, the user
        # is prompted to select from the list of values, but she presses
        # enter. The old value is returned.
        old_value = "bat"
        self.mocked_raw_input.side_effect = ["\n"]
        obtained = self.single_choice_param.prompt("", old_value=old_value)
        self.assertEquals(old_value, obtained)

    def test_without_old_value(self):
        # There is no old value, user just select the first choice.
        self.mocked_raw_input.side_effect = ["1"]
        obtained = self.single_choice_param.prompt("")
        self.assertEquals("foo", obtained)

    def test_with_wrong_user_input(self):
        # No old value, user inserts at least two wrong choices, and the select
        # the third one.
        self.mocked_raw_input.side_effect = ["1000", "0", "3"]
        obtained = self.single_choice_param.prompt("")
        self.assertEquals("baz", obtained)



# Copyright (C) 2013 Linaro Limited
#
# Author: Antonio Terceiro <antonio.terceiro@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

from lava.parameter import (
    ListParameter,
    Parameter,
)

LAVA_TEST_SHELL_TAR_REPO_KEY = "tar-repo"
LAVA_TEST_SHELL_TESDEF_KEY = "testdef"

DEVICE_TYPE_PARAMETER = Parameter("device_type")
PREBUILT_IMAGE_PARAMETER = Parameter("image", depends=DEVICE_TYPE_PARAMETER)

TESTDEF_URLS_PARAMETER = ListParameter("testdef_urls")
TESTDEF_URLS_PARAMETER.store = False

BOOT_TEST = {
    "timeout": 18000,
    "job_name": "Boot test",
    "device_type": DEVICE_TYPE_PARAMETER,
    "actions": [
        {
            "command": "deploy_linaro_image",
            "parameters": {
                "image": PREBUILT_IMAGE_PARAMETER
            }
        },
        {
            "command": "boot_linaro_image"
        }
    ]
}

LAVA_TEST_SHELL = {
    "job_name": "LAVA Test Shell",
    "timeout": 18000,
    "device_type": DEVICE_TYPE_PARAMETER,
    "actions": [
        {
            "command": "deploy_linaro_image",
            "parameters": {
                "image": PREBUILT_IMAGE_PARAMETER,
            }
        },
        {
            "command": "lava_test_shell",
            "parameters": {
                "timeout": 1800,
                "testdef_urls": TESTDEF_URLS_PARAMETER,
            }
        }
    ]
}

# This is a special case template, only use when automatically create job files
# starting from a testdef or a script. Never to be used directly by the user.
LAVA_TEST_SHELL_TAR_REPO = {
    "job_name": "LAVA Test Shell",
    "timeout": 18000,
    "device_type": DEVICE_TYPE_PARAMETER,
    "actions": [
        {
            "command": "deploy_linaro_image",
            "parameters": {
                "image": PREBUILT_IMAGE_PARAMETER,
            }
        },
        {
            "command": "lava_test_shell",
            "parameters": {
                "timeout": 1800,
                "testdef_repos": [
                    {
                        LAVA_TEST_SHELL_TESDEF_KEY: None,
                        LAVA_TEST_SHELL_TAR_REPO_KEY: None,
                    }
                ]
            }
        }
    ]
}

BOOT_TEST_KEY = "boot-test"
LAVA_TEST_SHELL_KEY = "lava-test-shell"

# Dict with all the user available job templates.
JOB_TYPES = {
    BOOT_TEST_KEY: BOOT_TEST,
    LAVA_TEST_SHELL_KEY: LAVA_TEST_SHELL,
}



# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""
Test definition commands class.
"""

import os
import tempfile

from lava.helper.command import BaseCommand
from lava.job import DEFAULT_JOB_FILENAME
from lava.tool.command import CommandGroup
from lava_tool.utils import verify_path_non_existance


class testdef(CommandGroup):

    """LAVA test definitions handling."""

    namespace = "lava.testdef.commands"


class TestdefBaseCommand(BaseCommand):

    def _create_tmp_job_file(self, testdef_file):
        testdef_file = os.path.abspath(testdef_file)
        verify_path_non_existance(testdef_file)

        job_file = os.path.join(tempfile.gettempdir(),
                                DEFAULT_JOB_FILENAME)

        tar_content = [testdef_file]
        job_file = self.create_tar_repo_job(job_file, testdef_file,
                                            tar_content)

        return job_file


class new(TestdefBaseCommand):

    """Creates a new test definition file."""

    @classmethod
    def register_arguments(cls, parser):
        super(new, cls).register_arguments(parser)
        parser.add_argument("FILE", help="Test definition file to create.")

    def invoke(self):
        full_path = os.path.abspath(self.args.FILE)
        self.create_test_definition(full_path)


class run(TestdefBaseCommand):

    """Runs the specified test definition on a local device."""

    @classmethod
    def register_arguments(cls, parser):
        super(run, cls).register_arguments(parser)
        parser.add_argument("FILE", help="Test definition file to run.")

    def invoke(self):
        job_file = ""
        try:
            job_file = self._create_tmp_job_file(self.args.FILE)
            super(run, self).run(job_file)
        finally:
            if os.path.isfile(job_file):
                os.unlink(job_file)


class submit(TestdefBaseCommand):

    """Submits the specified test definition to a LAVA server."""

    @classmethod
    def register_arguments(cls, parser):
        super(submit, cls).register_arguments(parser)
        parser.add_argument("FILE", help="Test definition file to send.")

    def invoke(self):
        job_file = ""
        try:
            job_file = self._create_tmp_job_file(self.args.FILE)
            super(submit, self).submit(job_file)
        finally:
            if os.path.isfile(job_file):
                os.unlink(job_file)



# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""lava.helper.dispatcher tests."""

import os
import tempfile

from mock import patch

from lava.tool.errors import CommandError
from lava.helper.tests.helper_test import HelperTest
from lava.helper.dispatcher import (
    choose_devices_path,
)


class DispatcherTests(HelperTest):

    def setUp(self):
        super(DispatcherTests, self).setUp()
        self.devices_dir = os.path.join(tempfile.gettempdir(), "devices")
        os.makedirs(self.devices_dir)

    def tearDown(self):
        super(DispatcherTests, self).tearDown()
        os.removedirs(self.devices_dir)

    def test_choose_devices_path_0(self):
        # Tests that when passing more than one path, the first writable one
        # is returned.
        obtained = choose_devices_path(
            ["/", "/root", self.temp_dir, os.path.expanduser("~")])
        expected = os.path.join(self.temp_dir, "devices")
        self.assertEqual(expected, obtained)

    def test_choose_devices_path_1(self):
        # Tests that when passing a path that is not writable, CommandError
        # is raised.
        self.assertRaises(CommandError, choose_devices_path,
                          ["/", "/root", "/root/tmpdir"])

    def test_choose_devices_path_2(self):
        # Tests that the correct path for devices is created on the filesystem.
        expected_path = os.path.join(self.temp_dir, "devices")
        obtained = choose_devices_path([self.temp_dir])
        self.assertEqual(expected_path, obtained)
        self.assertTrue(os.path.isdir(expected_path))

    def test_choose_devices_path_3(self):
        # Tests that returns the already existing devices path.
        obtained = choose_devices_path([tempfile.gettempdir()])
        self.assertEqual(self.devices_dir, obtained)

    @patch("__builtin__.open")
    def test_choose_devices_path_4(self, mocked_open):
        # Tests that when IOError is raised and we pass only one dir
        # CommandError is raised.
        mocked_open.side_effect = IOError()
        self.assertRaises(CommandError, choose_devices_path,
                          [tempfile.gettempdir()])
        self.assertTrue(mocked_open.called)



# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""
Tests for lava.testdef.commands.
"""

import os
import tempfile
import yaml

from mock import (
    MagicMock,
    patch,
)

from lava.config import InteractiveCache
from lava.helper.tests.helper_test import HelperTest
from lava.testdef.commands import (
    new,
)
from lava.tool.errors import CommandError


class NewCommandTest(HelperTest):

    """Class for the lava.testdef new command tests."""

    def setUp(self):
        super(NewCommandTest, self).setUp()
        self.file_name = "fake_testdef.yaml"
        self.file_path = os.path.join(tempfile.gettempdir(), self.file_name)
        self.args.FILE = self.file_path

        self.temp_yaml = tempfile.NamedTemporaryFile(suffix=".yaml",
                                                     delete=False)

        self.config_file = tempfile.NamedTemporaryFile(delete=False)
        self.config = InteractiveCache()
        self.config.save = MagicMock()
        self.config.config_file = self.config_file.name
        # Patch class raw_input, start it, and stop it on tearDown.
        self.patcher1 = patch("lava.parameter.raw_input", create=True)
        self.mocked_raw_input = self.patcher1.start()

    def tearDown(self):
        super(NewCommandTest, self).tearDown()
        if os.path.isfile(self.file_path):
            os.unlink(self.file_path)
        os.unlink(self.config_file.name)
        os.unlink(self.temp_yaml.name)
        self.patcher1.stop()

    def test_register_arguments(self):
        # Make sure that the parser add_argument is called and we have the
        # correct argument.
        new_command = new(self.parser, self.args)
        new_command.register_arguments(self.parser)

        # Make sure we do not forget about this test.
        self.assertEqual(2, len(self.parser.method_calls))

        _, args, _ = self.parser.method_calls[0]
        self.assertIn("--non-interactive", args)

        _, args, _ = self.parser.method_calls[1]
        self.assertIn("FILE", args)

    def test_invoke_0(self):
        # Test that passing a file on the command line, it is created on the
        # file system.
        self.mocked_raw_input.return_value = "\n"
        new_command = new(self.parser, self.args)
        new_command.config = self.config
        new_command.invoke()
        self.assertTrue(os.path.exists(self.file_path))

    def test_invoke_1(self):
        # Test that when passing an already existing file, an exception is
        # thrown.
        self.args.FILE = self.temp_yaml.name
        new_command = new(self.parser, self.args)
        new_command.config = self.config
        self.assertRaises(CommandError, new_command.invoke)

    def test_invoke_2(self):
        # Tests that when adding a new test definition and writing it to file
        # a correct YAML structure is created.
        self.mocked_raw_input.return_value = "\n"
        new_command = new(self.parser, self.args)
        new_command.config = self.config
        new_command.invoke()
        expected = {'run': {'steps': ["./mytest.sh"]},
                    'metadata': {
                        'environment': ['lava_test_shell'],
                        'format': 'Lava-Test Test Definition 1.0',
                        'version': '1.0',
                        'description': '',
                        'name': ''},
                    'parse': {
                        'pattern':
                        '^\\s*(?P<test_case_id>\\w+)=(?P<result>\\w+)\\s*$'
                    },
                    }
        obtained = None
        with open(self.file_path, 'r') as read_file:
            obtained = yaml.load(read_file)
        self.assertEqual(expected, obtained)

    def test_invoke_3(self):
        # Tests that when adding a new test definition and writing it to a file
        # in a directory withour permissions, exception is raised.
        self.args.FILE = "/test_file.yaml"
        self.mocked_raw_input.return_value = "\n"
        new_command = new(self.parser, self.args)
        new_command.config = self.config
        self.assertRaises(CommandError, new_command.invoke)
        self.assertFalse(os.path.exists(self.args.FILE))

    def test_invoke_4(self):
        # Tests that when passing values for the "steps" ListParameter, we get
        # back the correct data structure.
        self.mocked_raw_input.side_effect = ["foo", "\n", "\n", "\n", "\n",
                                             "\n"]
        new_command = new(self.parser, self.args)
        new_command.config = self.config
        new_command.invoke()
        expected = {'run': {'steps': ["./mytest.sh"]},
                    'metadata': {
                        'environment': ['lava_test_shell'],
                        'format': 'Lava-Test Test Definition 1.0',
                        'version': '1.0',
                        'description': '',
                        'name': 'foo'
                    },
                    'parse': {
                        'pattern':
                        '^\\s*(?P<test_case_id>\\w+)=(?P<result>\\w+)\\s*$'
                    },
                    }
        obtained = None
        with open(self.file_path, 'r') as read_file:
            obtained = yaml.load(read_file)
        self.assertEqual(expected, obtained)



# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""Scripts handling class."""

import os
import stat

from lava_tool.utils import write_file


DEFAULT_MOD = stat.S_IRWXU | stat.S_IRWXG | stat.S_IROTH | stat.S_IXOTH
DEFAULT_TESTDEF_SCRIPT_CONTENT = """#!/bin/sh
# Automatic generated content by lava-tool.
# Please add your own instructions.
#
# You can use all the avialable Bash commands.
#
# For the available LAVA commands, see:
#    http://lava.readthedocs.org/
#
"""
DEFAULT_TESTDEF_SCRIPT = "mytest.sh"


class ShellScript(object):

    """Creates a shell script on the file system with some content."""

    def __init__(self, file_name):
        self.file_name = file_name

    def write(self):
        write_file(self.file_name, DEFAULT_TESTDEF_SCRIPT_CONTENT)
        # Make sure the script is executable.
        os.chmod(self.file_name, DEFAULT_MOD)



# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""Base command class common to lava commands series."""

import os
import sys
import xmlrpclib

from lava.config import (
    InteractiveCache,
)
from lava.helper.dispatcher import get_devices
from lava.job import Job
from lava.job.templates import (
    LAVA_TEST_SHELL_TAR_REPO,
    LAVA_TEST_SHELL_TAR_REPO_KEY,
    LAVA_TEST_SHELL_TESDEF_KEY,
)
from lava.parameter import (
    Parameter,
    SingleChoiceParameter,
)
from lava.script import (
    ShellScript,
    DEFAULT_TESTDEF_SCRIPT,
)
from lava.testdef import TestDefinition
from lava.testdef.templates import (
    TESTDEF_STEPS_KEY,
    TESTDEF_TEMPLATE,
)
from lava.tool.command import Command
from lava.tool.errors import CommandError
from lava_tool.authtoken import (
    AuthenticatingServerProxy,
    KeyringAuthBackend
)
from lava_tool.utils import (
    base64_encode,
    create_tar,
    execute,
    has_command,
    to_list,
    verify_and_create_url,
)

CONFIG = InteractiveCache()


class BaseCommand(Command):

    """Base command class for all lava commands."""

    def __init__(self, parser, args):
        super(BaseCommand, self).__init__(parser, args)
        self.config = CONFIG
        self.config.force_interactive = self.args.non_interactive

    @classmethod
    def register_arguments(cls, parser):
        super(BaseCommand, cls).register_arguments(parser)
        parser.add_argument("--non-interactive", "-n",
                            action='store_false',
                            help=("Do not ask for input parameters."))

    def authenticated_server(self):
        """Returns a connection to a LAVA server.

        It will ask the user the necessary parameters to establish the
        connection.
        """
        print >> sys.stdout, "\nServer connection parameters:"

        server_name_parameter = Parameter("server")
        rpc_endpoint_parameter = Parameter("rpc_endpoint",
                                           depends=server_name_parameter)

        self.config.get(server_name_parameter)
        endpoint = self.config.get(rpc_endpoint_parameter)

        rpc_url = verify_and_create_url(endpoint)
        server = AuthenticatingServerProxy(rpc_url,
                                           auth_backend=KeyringAuthBackend())
        return server

    def submit(self, job_file):
        """Submits a job file to a LAVA server.

        :param job_file: The job file to submit.
        :return The job ID on success.
        """
        if os.path.isfile(job_file):
            try:
                jobdata = open(job_file, 'rb').read()
                server = self.authenticated_server()

                job_id = server.scheduler.submit_job(jobdata)
                print >> sys.stdout, ("Job submitted with job "
                                      "ID {0}.".format(job_id))

                return job_id
            except xmlrpclib.Fault, exc:
                raise CommandError(str(exc))
        else:
            raise CommandError("Job file '{0}' does not exists, or is not "
                               "a file.".format(job_file))

    def run(self, job_file):
        """Runs a job file on the local LAVA dispatcher.

        :param job_file: The job file to run.
        """
        if os.path.isfile(job_file):
            if has_command("lava-dispatch"):
                devices = get_devices()
                if devices:
                    if len(devices) > 1:
                        device_names = [device.hostname for device in devices]
                        device_param = SingleChoiceParameter("device",
                                                             device_names)
                        device = device_param.prompt("Device to use: ")
                    else:
                        device = devices[0].hostname
                    execute(["lava-dispatch", "--target", device, job_file])
            else:
                raise CommandError("Cannot find lava-dispatcher installation.")
        else:
            raise CommandError("Job file '{0}' does not exists, or it is not "
                               "a file.".format(job_file))

    def status(self, job_id):
        """Retrieves the status of a LAVA job.

        :param job_id: The ID of the job to look up.
        """
        job_id = str(job_id)

        try:
            server = self.authenticated_server()
            job_status = server.scheduler.job_status(job_id)

            status = job_status["job_status"].lower()
            bundle = job_status["bundle_sha1"]

            print >> sys.stdout, "\nJob id: {0}".format(job_id)
            print >> sys.stdout, "Status: {0}".format(status)
            print >> sys.stdout, "Bundle: {0}".format(bundle)
        except xmlrpclib.Fault, exc:
            raise CommandError(str(exc))

    def details(self, job_id):
        """Retrieves the details of a LAVA job.

        :param job_id: The ID of the job to look up.
        """
        job_id = str(job_id)

        try:
            server = self.authenticated_server()
            job_details = server.scheduler.job_details(job_id)

            print >> sys.stdout, "\nDetails of job {0}:\n".format(job_id)
            print >> sys.stdout, job_details
        except xmlrpclib.Fault, exc:
            raise CommandError(str(exc))

    def create_tar_repo_job(self, job_file, testdef_file, tar_content):
        """Creates a job file based on the tar-repo template.

        The tar repo is not kept on the file system.

        :param job_file: The path of the job file to create.
        :param testdef_file: The path of the test definition file.
        :param tar_content: What should go into the tarball repository.
        :return The path of the job file created.
        """

        print >> sys.stdout, "\nCreating job file..."

        try:
            tar_repo = create_tar(tar_content)

            job_instance = Job(LAVA_TEST_SHELL_TAR_REPO, job_file)
            job_instance.update(self.config)

            job_instance.set(LAVA_TEST_SHELL_TAR_REPO_KEY,
                             base64_encode(tar_repo))
            job_instance.set(LAVA_TEST_SHELL_TESDEF_KEY,
                             os.path.basename(testdef_file))

            job_instance.write()

            basename = os.path.basename(job_instance.file_name)
            print >> sys.stdout, ("\nCreated job file "
                                  "'{0}'.".format(basename))

            return job_instance.file_name
        finally:
            if os.path.isfile(tar_repo):
                os.unlink(tar_repo)

    def create_test_definition(self, testdef_file, template=TESTDEF_TEMPLATE,
                               steps=None):
        """Creates a test definition YAML file.

        :param testdef_file: The file to create.
        :return The path of the file created.
        """

        print >> sys.stdout, "\nCreating test definition file..."

        testdef = TestDefinition(template, testdef_file)
        if steps:
            steps = to_list(steps)
            testdef.set(TESTDEF_STEPS_KEY, steps)
        testdef.update(self.config)
        testdef.write()

        basename = os.path.basename(testdef.file_name)
        print >> sys.stdout, ("\nCreated test definition "
                              "'{0}'.".format(basename))

        return testdef.file_name

    def create_shell_script(self, test_path,
                            script_name=DEFAULT_TESTDEF_SCRIPT):
        """Creates a shell script with some default content.

        :param test_path: The directory where to create the script.
        :param script_name: The name of the script.
        :return The full path to the script file.
        """
        default_script = os.path.join(test_path, script_name)

        if not os.path.isfile(default_script):
            print >> sys.stdout, "Creating shell script..."

            shell_script = ShellScript(default_script)
            shell_script.write()

            print >> sys.stdout, ("\nCreated shell script "
                                  "'{0}'.".format(script_name))

        return default_script



# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""
Tests for lava.script.commands.
"""

from lava.helper.tests.helper_test import HelperTest
from lava.script.commands import (
    run,
    submit,
)


class RunCommandTests(HelperTest):

    def test_register_arguments(self):
        run_cmd = run(self.parser, self.args)
        run_cmd.register_arguments(self.parser)

        # Make sure we do not forget about this test.
        self.assertEqual(2, len(self.parser.method_calls))

        _, args, _ = self.parser.method_calls[0]
        self.assertIn("--non-interactive", args)

        _, args, _ = self.parser.method_calls[1]
        self.assertIn("FILE", args)


class SubmitCommandTests(HelperTest):

    def test_register_arguments(self):
        submit_cmd = submit(self.parser, self.args)
        submit_cmd.register_arguments(self.parser)

        # Make sure we do not forget about this test.
        self.assertEqual(2, len(self.parser.method_calls))

        _, args, _ = self.parser.method_calls[0]
        self.assertIn("--non-interactive", args)

        _, args, _ = self.parser.method_calls[1]
        self.assertIn("FILE", args)



# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""Helper functions for a template."""

from lava.parameter import Parameter


def expand_template(template, config):
    """Updates a template based on the values from the provided config.

    :param template: A template to be updated.
    :param config: A Config instance where values should be taken.
    """

    def update(data):
        """Internal recursive function."""
        if isinstance(data, dict):
            keys = data.keys()
        elif isinstance(data, list):
            keys = range(len(data))
        else:
            return
        for key in keys:
            entry = data[key]
            if isinstance(entry, Parameter):
                data[key] = config.get(entry)
            else:
                update(entry)

    update(template)


def get_key(data, search_key):
    """Goes through a template looking for a key.

    :param data: The template to traverse.
    :param search_key: The key to look for.
    :return The key value.
    """
    return_value = None
    found = False

    if isinstance(data, dict):
        bucket = []

        for key, value in data.iteritems():
            if key == search_key:
                return_value = value
                found = True
                break
            else:
                bucket.append(value)

        if bucket and not found:
            for element in bucket:
                if isinstance(element, list):
                    for element in element:
                        bucket.append(element)
                elif isinstance(element, dict):
                    for key, value in element.iteritems():
                        if key == search_key:
                            return_value = value
                            found = True
                            break
                        else:
                            bucket.append(value)
                if found:
                    break

    return return_value


def set_value(data, search_key, new_value):
    """Sets a new value for a template key.

    :param data: The data structure to update.
    :type dict
    :param search_key: The key to search and update.
    :param new_value: The new value to set.
    """
    is_set = False

    if isinstance(data, dict):
        bucket = []

        for key, value in data.iteritems():
            if key == search_key:
                data[key] = new_value
                is_set = True
                break
            else:
                bucket.append(value)

        if bucket and not is_set:
            for element in bucket:
                if isinstance(element, list):
                    for element in element:
                        bucket.append(element)
                elif isinstance(element, dict):
                    for key, value in element.iteritems():
                        if key == search_key:
                            element[key] = new_value
                            is_set = True
                            break
                        else:
                            bucket.append(value)
                if is_set:
                    break



# Copyright (C) 2010,2011 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of lava-dashboard-tool.
#
# lava-dashboard-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-dashboard-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-dashboard-tool.  If not, see <http://www.gnu.org/licenses/>.

"""
Module with command-line tool commands that interact with the dashboard
server. All commands listed here should have counterparts in the
launch_control.dashboard_app.xml_rpc package.
"""

import argparse
import contextlib
import errno
import os
import re
import socket
import sys
import urllib
import urlparse
import xmlrpclib

import simplejson
from json_schema_validator.extensions import datetime_extension

from lava_tool.authtoken import AuthenticatingServerProxy, KeyringAuthBackend
from lava.tool.command import Command, CommandGroup


class dashboard(CommandGroup):

    """
    Commands for interacting with LAVA Dashboard
    """

    namespace = "lava.dashboard.commands"


class InsufficientServerVersion(Exception):

    """
    Exception raised when server version that a command interacts with is too
    old to support required features.
    """

    def __init__(self, server_version, required_version):
        self.server_version = server_version
        self.required_version = required_version


class DataSetRenderer(object):

    """
    Support class for rendering a table out of list of dictionaries.

    It supports several features that make printing tabular data easier.
    * Automatic layout
    * Custom column headers
    * Custom cell formatting
    * Custom table captions
    * Custom column ordering
    * Custom Column separators
    * Custom dataset notification

    The primary method is render() which does all of the above. You
    need to pass a dataset argument which is a list of dictionaries.
    Each dictionary must have the same keys. In particular the first row
    is used to determine columns.
    """

    def __init__(self, column_map=None, row_formatter=None, empty=None,
                 order=None, caption=None, separator=" ",
                 header_separator=None):
        if column_map is None:
            column_map = {}
        if row_formatter is None:
            row_formatter = {}
        if empty is None:
            empty = "There is no data to display"
        self.column_map = column_map
        self.row_formatter = row_formatter
        self.empty = empty
        self.order = order
        self.separator = separator
        self.caption = caption
        self.header_separator = header_separator

    def _analyze_dataset(self, dataset):
        """
        Determine the columns that will be displayed and the maximum
        length of each of those columns.

        Returns a tuple (dataset, columms, maxlen) where columns is a
        list of column names and maxlen is a dictionary mapping from
        column name to maximum length of any value in the row or the
        column header and the dataset is a copy of the dataset altered
        as necessary.

        Some examples:

        First the dataset, an array of dictionaries
        >>> dataset = [
        ...     {'a': 'shorter', 'bee': ''},
        ...     {'a': 'little longer', 'bee': 'b'}]

        Note that column 'bee' is actually three characters long as the
        column name made it wider.
        >>> dataset_out, columns, maxlen = DataSetRenderer(
        ...     )._analyze_dataset(dataset)

        Unless you format rows with a custom function the data is not altered.
        >>> dataset_out is dataset
        True

        Columns come out in sorted alphabetic order
        >>> columns
        ['a', 'bee']

        Maximum length determines the width of each column. Note that
        the header affects the column width.
        >>> maxlen
        {'a': 13, 'bee': 3}

        You can constrain or reorder columns. In that case columns you
        decided to ignore are simply left out of the output.
        >>> dataset_out, columns, maxlen = DataSetRenderer(
        ...     order=['bee'])._analyze_dataset(dataset)
        >>> columns
        ['bee']
        >>> maxlen
        {'bee': 3}

        You can format values anyway you like:
        >>> dataset_out, columns, maxlen = DataSetRenderer(row_formatter={
        ...     'bee': lambda value: "%10s" % value}
        ...     )._analyze_dataset(dataset)

        Dataset is altered to take account of the row formatting
        function. The original dataset argument is copied.
        >>> dataset_out
        [{'a': 'shorter', 'bee': '          '}, {'a': 'little longer', 'bee': '         b'}]
        >>> dataset_out is not dataset
        True

        Columns stay the same though:
        >>> columns
        ['a', 'bee']

        Note how formatting altered the width of the column 'bee'
        >>> maxlen
        {'a': 13, 'bee': 10}

        You can also format columns (with nice aliases).Note how
        column 'bee' maximum width is now dominated by the long column
        name:
        >>> dataset_out, columns, maxlen = DataSetRenderer(column_map={
        ...     'bee': "Column B"})._analyze_dataset(dataset)
        >>> maxlen
        {'a': 13, 'bee': 8}
        """
        if self.order:
            columns = self.order
        else:
            columns = sorted(dataset[0].keys())
        if self.row_formatter:
            dataset_out = [dict(row) for row in dataset]
        else:
            dataset_out = dataset
        for row in dataset_out:
            for column in row:
                if column in self.row_formatter:
                    row[column] = self.row_formatter[column](row[column])
        maxlen = dict(
            [(column, max(
              len(self.column_map.get(column, column)),
              max([
                  len(str(row[column])) for row in dataset_out])))
             for column in columns])
        return dataset_out, columns, maxlen

    def _render_header(self, dataset, columns, maxlen):
        """
        Render a header, possibly with a caption string

        Caption is controlled by the constructor.
        >>> dataset = [
        ...     {'a': 'shorter', 'bee': ''},
        ...     {'a': 'little longer', 'bee': 'b'}]
        >>> columns = ['a', 'bee']
        >>> maxlen = {'a': 13, 'bee': 3}

        By default there is no caption, just column names:
        >>> DataSetRenderer()._render_header(
        ...     dataset, columns, maxlen)
              a       bee

        If you enable the header separator then column names will be visually
        separated from the first row of data.
        >>> DataSetRenderer(header_separator=True)._render_header(
        ...     dataset, columns, maxlen)
              a       bee
        -----------------

        If you provide a caption it gets rendered as a centered
        underlined text before the data:
        >>> DataSetRenderer(caption="Dataset")._render_header(
        ...     dataset, columns, maxlen)
             Dataset     
        =================
              a       bee

        You can use both caption and header separator
        >>> DataSetRenderer(caption="Dataset", header_separator=True)._render_header(
        ...     dataset, columns, maxlen)
             Dataset     
        =================
              a       bee
        -----------------

        Observe how the total length of the output horizontal line
        depends on the separator! Also note the columns labels are
        aligned to the center of the column
        >>> DataSetRenderer(caption="Dataset", separator=" | ")._render_header(
        ...     dataset, columns, maxlen)
              Dataset      
        ===================
              a       | bee
        """
        total_len = sum(maxlen.itervalues())
        if len(columns):
            total_len += len(self.separator) * (len(columns) - 1)
        # Print the caption
        if self.caption:
            print "{0:^{1}}".format(self.caption, total_len)
            print "=" * total_len
        # Now print the coulum names
        print self.separator.join([
            "{0:^{1}}".format(self.column_map.get(column, column),
                              maxlen[column]) for column in columns])
        # Finally print the header separator
        if self.header_separator:
            print "-" * total_len

    def _render_rows(self, dataset, columns, maxlen):
        """
        Render rows of the dataset.

        Each row is printed on one line using the maxlen argument to
        determine correct column size. Text is aligned left.

        First the dataset, columns and maxlen as produced by
        _analyze_dataset()
        >>> dataset = [
        ...     {'a': 'shorter', 'bee': ''},
        ...     {'a': 'little longer', 'bee': 'b'}]
        >>> columns = ['a', 'bee']
        >>> maxlen = {'a': 13, 'bee': 3}

        Now a plain table. Note! To really understand this test
        you should check out the whitespace in the strings below. There
        are two more spaces after 'b' in the second row
        >>> DataSetRenderer()._render_rows(dataset, columns, maxlen)
        shorter          
        little longer b  
        """
        for row in dataset:
            print self.separator.join([
                "{0!s:{1}}".format(row[column], maxlen[column])
                for column in columns])

    def _render_dataset(self, dataset):
        """
        Render the header followed by the rows of data.
        """
        dataset, columns, maxlen = self._analyze_dataset(dataset)
        self._render_header(dataset, columns, maxlen)
        self._render_rows(dataset, columns, maxlen)

    def _render_empty_dataset(self):
        """
        Render empty dataset.

        By default it just prints out a fixed sentence:
        >>> DataSetRenderer()._render_empty_dataset()
        There is no data to display

        This can be changed by passing an argument to the constructor
        >>> DataSetRenderer(empty="there is no data")._render_empty_dataset()
        there is no data
        """
        print self.empty

    def render(self, dataset):
        if len(dataset) > 0:
            self._render_dataset(dataset)
        else:
            self._render_empty_dataset()


class XMLRPCCommand(Command):

    """
    Abstract base class for commands that interact with dashboard server
    over XML-RPC.

    The only difference is that you should implement invoke_remote()
    instead of invoke(). The provided implementation catches several
    socket and XML-RPC errors and prints a pretty error message.
    """

    @staticmethod
    def _construct_xml_rpc_url(url):
        """
        Construct URL to the XML-RPC service out of the given URL
        """
        parts = urlparse.urlsplit(url)
        if not parts.path.endswith("/RPC2/"):
            path = parts.path.rstrip("/") + "/xml-rpc/"
        else:
            path = parts.path
        return urlparse.urlunsplit(
            (parts.scheme, parts.netloc, path, "", ""))

    @staticmethod
    def _strict_server_version(version):
        """
        Calculate strict server version (as defined by
        distutils.version.StrictVersion). This works by discarding .candidate
        and .dev release-levels.
        >>> XMLRPCCommand._strict_server_version("0.4.0.candidate.5")
        '0.4.0'
        >>> XMLRPCCommand._strict_server_version("0.4.0.dev.126")
        '0.4.0'
        >>> XMLRPCCommand._strict_server_version("0.4.0.alpha.1")
        '0.4.0a1'
        >>> XMLRPCCommand._strict_server_version("0.4.0.beta.2")
        '0.4.0b2'
        """
        try:
            major, minor, micro, releaselevel, serial = version.split(".")
        except ValueError:
            raise ValueError(
                ("version %r does not follow pattern "
                 "'major.minor.micro.releaselevel.serial'") % version)
        if releaselevel in ["dev", "candidate", "final"]:
            return "%s.%s.%s" % (major, minor, micro)
        elif releaselevel == "alpha":
            return "%s.%s.%sa%s" % (major, minor, micro, serial)
        elif releaselevel == "beta":
            return "%s.%s.%sb%s" % (major, minor, micro, serial)
        else:
            raise ValueError(
                ("releaselevel %r is not one of 'final', 'alpha', 'beta', "
                 "'candidate' or 'final'") % releaselevel)

    def _check_server_version(self, server_obj, required_version):
        """
        Check that server object has is at least required_version.

        This method may raise InsufficientServerVersion.
        """
        from distutils.version import StrictVersion, LooseVersion
        # For backwards compatibility the server reports
        # major.minor.micro.releaselevel.serial which is not PEP-386 compliant
        server_version = StrictVersion(
            self._strict_server_version(server_obj.version()))
        required_version = StrictVersion(required_version)
        if server_version < required_version:
            raise InsufficientServerVersion(server_version, required_version)

    def __init__(self, parser, args):
        super(XMLRPCCommand, self).__init__(parser, args)
        xml_rpc_url = self._construct_xml_rpc_url(self.args.dashboard_url)
        self.server = AuthenticatingServerProxy(
            xml_rpc_url,
            verbose=args.verbose_xml_rpc,
            allow_none=True,
            use_datetime=True,
            auth_backend=KeyringAuthBackend())

    def use_non_legacy_api_if_possible(self, name='server'):
        # Legacy APIs are registered in top-level object, non-legacy APIs are
        # prefixed with extension name.
        if "dashboard.version" in getattr(self, name).system.listMethods():
            setattr(self, name, getattr(self, name).dashboard)

    @classmethod
    def register_arguments(cls, parser):
        dashboard_group = parser.add_argument_group(
            "dashboard specific arguments")
        default_dashboard_url = os.getenv("DASHBOARD_URL")
        if default_dashboard_url:
            dashboard_group.add_argument("--dashboard-url",
                                         metavar="URL", help="URL of your validation dashboard (currently %(default)s)",
                                         default=default_dashboard_url)
        else:
            dashboard_group.add_argument("--dashboard-url", required=True,
                                         metavar="URL", help="URL of your validation dashboard")
        debug_group = parser.add_argument_group("debugging arguments")
        debug_group.add_argument("--verbose-xml-rpc",
                                 action="store_true", default=False,
                                 help="Show XML-RPC data")
        return dashboard_group

    @contextlib.contextmanager
    def safety_net(self):
        try:
            yield
        except socket.error as ex:
            print >> sys.stderr, "Unable to connect to server at %s" % (
                self.args.dashboard_url,)
            # It seems that some errors are reported as -errno
            # while others as +errno.
            ex.errno = abs(ex.errno)
            if ex.errno == errno.ECONNREFUSED:
                print >> sys.stderr, "Connection was refused."
                parts = urlparse.urlsplit(self.args.dashboard_url)
                if parts.netloc == "localhost:8000":
                    print >> sys.stderr, "Perhaps the server is not running?"
            elif ex.errno == errno.ENOENT:
                print >> sys.stderr, "Unable to resolve address"
            else:
                print >> sys.stderr, "Socket %d: %s" % (ex.errno, ex.strerror)
        except xmlrpclib.ProtocolError as ex:
            print >> sys.stderr, "Unable to exchange XML-RPC message with dashboard server"
            print >> sys.stderr, "HTTP error code: %d/%s" % (
                ex.errcode, ex.errmsg)
        except xmlrpclib.Fault as ex:
            self.handle_xmlrpc_fault(ex.faultCode, ex.faultString)
        except InsufficientServerVersion as ex:
            print >> sys.stderr, ("This command requires at least server version "
                                  "%s, actual server version is %s" %
                                 (ex.required_version, ex.server_version))

    def invoke(self):
        with self.safety_net():
            self.use_non_legacy_api_if_possible()
            return self.invoke_remote()

    def handle_xmlrpc_fault(self, faultCode, faultString):
        if faultCode == 500:
            print >> sys.stderr, "Dashboard server has experienced internal error"
            print >> sys.stderr, faultString
        else:
            print >> sys.stderr, "XML-RPC error %d: %s" % (
                faultCode, faultString)

    def invoke_remote(self):
        raise NotImplementedError()


class server_version(XMLRPCCommand):

    """
    Display dashboard server version
    """

    def invoke_remote(self):
        print "Dashboard server version: %s" % (self.server.version(),)


class put(XMLRPCCommand):

    """
    Upload a bundle on the server
    """

    @classmethod
    def register_arguments(cls, parser):
        super(put, cls).register_arguments(parser)
        parser.add_argument("LOCAL",
                            type=argparse.FileType("rb"),
                            help="pathname on the local file system")
        parser.add_argument("REMOTE",
                            default="/anonymous/", nargs='?',
                            help="pathname on the server")

    def invoke_remote(self):
        content = self.args.LOCAL.read()
        filename = self.args.LOCAL.name
        pathname = self.args.REMOTE
        content_sha1 = self.server.put(content, filename, pathname)
        print "Stored as bundle {0}".format(content_sha1)

    def handle_xmlrpc_fault(self, faultCode, faultString):
        if faultCode == 404:
            print >> sys.stderr, "Bundle stream %s does not exist" % (
                self.args.REMOTE)
        elif faultCode == 409:
            print >> sys.stderr, "You have already uploaded this bundle to the dashboard"
        else:
            super(put, self).handle_xmlrpc_fault(faultCode, faultString)


class get(XMLRPCCommand):

    """
    Download a bundle from the server
    """

    @classmethod
    def register_arguments(cls, parser):
        super(get, cls).register_arguments(parser)
        parser.add_argument("SHA1",
                            type=str,
                            help="SHA1 of the bundle to download")
        parser.add_argument("--overwrite",
                            action="store_true",
                            help="Overwrite files on the local disk")
        parser.add_argument("--output", "-o",
                            type=argparse.FileType("wb"),
                            default=None,
                            help="Alternate name of the output file")

    def invoke_remote(self):
        response = self.server.get(self.args.SHA1)
        if self.args.output is None:
            filename = self.args.SHA1
            if os.path.exists(filename) and not self.args.overwrite:
                print >> sys.stderr, "File {filename!r} already exists".format(
                    filename=filename)
                print >> sys.stderr, "You may pass --overwrite to write over it"
                return -1
            stream = open(filename, "wb")
        else:
            stream = self.args.output
            filename = self.args.output.name
        stream.write(response['content'])
        print "Downloaded bundle {0} to file {1!r}".format(
            self.args.SHA1, filename)

    def handle_xmlrpc_fault(self, faultCode, faultString):
        if faultCode == 404:
            print >> sys.stderr, "Bundle {sha1} does not exist".format(
                sha1=self.args.SHA1)
        else:
            super(get, self).handle_xmlrpc_fault(faultCode, faultString)


class deserialize(XMLRPCCommand):

    """
    Deserialize a bundle on the server
    """

    @classmethod
    def register_arguments(cls, parser):
        super(deserialize, cls).register_arguments(parser)
        parser.add_argument("SHA1",
                            type=str,
                            help="SHA1 of the bundle to deserialize")

    def invoke_remote(self):
        response = self.server.deserialize(self.args.SHA1)
        print "Bundle {sha1} deserialized".format(
            sha1=self.args.SHA1)

    def handle_xmlrpc_fault(self, faultCode, faultString):
        if faultCode == 404:
            print >> sys.stderr, "Bundle {sha1} does not exist".format(
                sha1=self.args.SHA1)
        elif faultCode == 409:
            print >> sys.stderr, "Unable to deserialize bundle {sha1}".format(
                sha1=self.args.SHA1)
            print >> sys.stderr, faultString
        else:
            super(
                deserialize,
                self).handle_xmlrpc_fault(
                faultCode,
                faultString)


def _get_pretty_renderer(**kwargs):
    if "separator" not in kwargs:
        kwargs["separator"] = " | "
    if "header_separator" not in kwargs:
        kwargs["header_separator"] = True
    return DataSetRenderer(**kwargs)


class streams(XMLRPCCommand):

    """
    Show streams you have access to
    """

    renderer = _get_pretty_renderer(
        order=('pathname', 'bundle_count', 'name'),
        column_map={
            'pathname': 'Pathname',
            'bundle_count': 'Number of bundles',
            'name': 'Name'},
        row_formatter={
            'name': lambda name: name or "(not set)"},
        empty="There are no streams you can access on the server",
        caption="Bundle streams")

    def invoke_remote(self):
        self.renderer.render(self.server.streams())


class bundles(XMLRPCCommand):

    """
    Show bundles in the specified stream
    """

    renderer = _get_pretty_renderer(
        column_map={
            'uploaded_by': 'Uploader',
            'uploaded_on': 'Upload date',
            'content_filename': 'File name',
            'content_sha1': 'SHA1',
            'is_deserialized': "Deserialized?"},
        row_formatter={
            'is_deserialized': lambda x: "yes" if x else "no",
            'uploaded_by': lambda x: x or "(anonymous)",
            'uploaded_on': lambda x: x},
        order=('content_sha1', 'content_filename', 'uploaded_by',
               'uploaded_on', 'is_deserialized'),
        empty="There are no bundles in this stream",
        caption="Bundles",
        separator=" | ")

    @classmethod
    def register_arguments(cls, parser):
        super(bundles, cls).register_arguments(parser)
        parser.add_argument("PATHNAME",
                            default="/anonymous/", nargs='?',
                            help="pathname on the server (defaults to %(default)s)")

    def invoke_remote(self):
        self.renderer.render(self.server.bundles(self.args.PATHNAME))

    def handle_xmlrpc_fault(self, faultCode, faultString):
        if faultCode == 404:
            print >> sys.stderr, "Bundle stream %s does not exist" % (
                                 self.args.PATHNAME)
        else:
            super(bundles, self).handle_xmlrpc_fault(faultCode, faultString)


class make_stream(XMLRPCCommand):

    """
    Create a bundle stream on the server
    """

    @classmethod
    def register_arguments(cls, parser):
        super(make_stream, cls).register_arguments(parser)
        parser.add_argument(
            "pathname",
            type=str,
            help="Pathname of the bundle stream to create")
        parser.add_argument(
            "--name",
            type=str,
            default="",
            help="Name of the bundle stream (description)")

    def invoke_remote(self):
        self._check_server_version(self.server, "0.3")
        pathname = self.server.make_stream(self.args.pathname, self.args.name)
        print "Bundle stream {pathname} created".format(pathname=pathname)


class backup(XMLRPCCommand):

    """
    Backup data uploaded to a dashboard instance.
    Not all data is preserved. The following data is lost: identity of the user
    that uploaded each bundle, time of uploading and deserialization on the
    server, name of the bundle stream that contained the data
    """

    @classmethod
    def register_arguments(cls, parser):
        super(backup, cls).register_arguments(parser)
        parser.add_argument("BACKUP_DIR", type=str,
                            help="Directory to backup to")

    def invoke_remote(self):
        if not os.path.exists(self.args.BACKUP_DIR):
            os.mkdir(self.args.BACKUP_DIR)
        for bundle_stream in self.server.streams():
            print "Processing stream %s" % bundle_stream["pathname"]
            bundle_stream_dir = os.path.join(
                self.args.BACKUP_DIR,
                urllib.quote_plus(bundle_stream["pathname"]))
            if not os.path.exists(bundle_stream_dir):
                os.mkdir(bundle_stream_dir)
            with open(os.path.join(bundle_stream_dir, "metadata.json"), "wt") as stream:
                simplejson.dump({
                    "pathname": bundle_stream["pathname"],
                    "name": bundle_stream["name"],
                    "user": bundle_stream["user"],
                    "group": bundle_stream["group"],
                }, stream)
            for bundle in self.server.bundles(bundle_stream["pathname"]):
                print " * Backing up bundle %s" % bundle["content_sha1"]
                data = self.server.get(bundle["content_sha1"])
                bundle_pathname = os.path.join(
                    bundle_stream_dir,
                    bundle["content_sha1"])
                # Note: we write bundles as binary data to preserve anything
                # the user might have dumped on us
                with open(bundle_pathname + ".json", "wb") as stream:
                    stream.write(data["content"])
                with open(bundle_pathname + ".metadata.json", "wt") as stream:
                    simplejson.dump({
                        "uploaded_by": bundle["uploaded_by"],
                        "uploaded_on":
                        datetime_extension.to_json(bundle["uploaded_on"]),
                        "content_filename": bundle["content_filename"],
                        "content_sha1": bundle["content_sha1"],
                        "content_size": bundle["content_size"],
                    }, stream)


class restore(XMLRPCCommand):

    """
    Restore a dashboard instance from backup
    """

    @classmethod
    def register_arguments(cls, parser):
        super(restore, cls).register_arguments(parser)
        parser.add_argument("BACKUP_DIR", type=str,
                            help="Directory to backup from")

    def invoke_remote(self):
        self._check_server_version(self.server, "0.3")
        for stream_pathname_quoted in os.listdir(self.args.BACKUP_DIR):
            filesystem_stream_pathname = os.path.join(
                self.args.BACKUP_DIR,
                stream_pathname_quoted)
            if not os.path.isdir(filesystem_stream_pathname):
                continue
            stream_pathname = urllib.unquote(stream_pathname_quoted)
            if os.path.exists(os.path.join(filesystem_stream_pathname, "metadata.json")):
                with open(os.path.join(filesystem_stream_pathname, "metadata.json"), "rt") as stream:
                    stream_metadata = simplejson.load(stream)
            else:
                stream_metadata = {}
            print "Processing stream %s" % stream_pathname
            try:
                self.server.make_stream(
                    stream_pathname,
                    stream_metadata.get(
                        "name",
                        "Restored from backup"))
            except xmlrpclib.Fault as ex:
                if ex.faultCode != 409:
                    raise
            for content_sha1 in [item[:-len(".json")] for item in os.listdir(filesystem_stream_pathname) if item.endswith(".json") and not item.endswith(".metadata.json") and item != "metadata.json"]:
                filesystem_content_filename = os.path.join(
                    filesystem_stream_pathname,
                    content_sha1 + ".json")
                if not os.path.isfile(filesystem_content_filename):
                    continue
                with open(os.path.join(filesystem_stream_pathname, content_sha1) + ".metadata.json", "rt") as stream:
                    bundle_metadata = simplejson.load(stream)
                with open(filesystem_content_filename, "rb") as stream:
                    content = stream.read()
                print " * Restoring bundle %s" % content_sha1
                try:
                    self.server.put(
                        content,
                        bundle_metadata["content_filename"],
                        stream_pathname)
                except xmlrpclib.Fault as ex:
                    if ex.faultCode != 409:
                        raise


class pull(XMLRPCCommand):
    """
    Copy bundles and bundle streams from one dashboard to another.
    This command checks for two environment varialbes:
    The value of DASHBOARD_URL is used as a replacement for --dashbard-url.
    The value of REMOTE_DASHBOARD_URL as a replacement for FROM.
    Their presence automatically makes the corresponding argument optional.
    """

    def __init__(self, parser, args):
        super(pull, self).__init__(parser, args)
        remote_xml_rpc_url = self._construct_xml_rpc_url(self.args.FROM)
        self.remote_server = AuthenticatingServerProxy(
            remote_xml_rpc_url,
            verbose=args.verbose_xml_rpc,
            use_datetime=True,
            allow_none=True,
            auth_backend=KeyringAuthBackend())
        self.use_non_legacy_api_if_possible('remote_server')

    @classmethod
    def register_arguments(cls, parser):
        group = super(pull, cls).register_arguments(parser)
        default_remote_dashboard_url = os.getenv("REMOTE_DASHBOARD_URL")
        if default_remote_dashboard_url:
            group.add_argument(
                "FROM", nargs="?",
                help="URL of the remote validation dashboard (currently %(default)s)",
                default=default_remote_dashboard_url)
        else:
            group.add_argument(
                "FROM",
                help="URL of the remote validation dashboard)")
        group.add_argument(
            "STREAM",
            nargs="*",
            help="Streams to pull from (all by default)")

    @staticmethod
    def _filesizeformat(num_bytes):
        """
        Formats the value like a 'human-readable' file size (i.e. 13 KB, 4.1 MB,
        102 num_bytes, etc).
        """
        try:
            num_bytes = float(num_bytes)
        except (TypeError, ValueError, UnicodeDecodeError):
            return "%(size)d byte", "%(size)d num_bytes" % {'size': 0}

        filesize_number_format = lambda value: "%0.2f" % (round(value, 1),)

        if num_bytes < 1024:
            return "%(size)d bytes" % {'size': num_bytes}
        if num_bytes < 1024 * 1024:
            return "%s KB" % filesize_number_format(num_bytes / 1024)
        if num_bytes < 1024 * 1024 * 1024:
            return "%s MB" % filesize_number_format(num_bytes / (1024 * 1024))
        return "%s GB" % filesize_number_format(num_bytes / (1024 * 1024 * 1024))

    def invoke_remote(self):
        self._check_server_version(self.server, "0.3")

        print "Checking local and remote streams"
        remote = self.remote_server.streams()
        if self.args.STREAM:
            # Check that all requested streams are available remotely
            requested_set = frozenset(self.args.STREAM)
            remote_set = frozenset((stream["pathname"] for stream in remote))
            unavailable_set = requested_set - remote_set
            if unavailable_set:
                print >> sys.stderr, "Remote stream not found: %s" % ", ".join(
                    unavailable_set)
                return -1
            # Limit to requested streams if necessary
            remote = [
                stream for stream in remote if stream[
                    "pathname"] in requested_set]
        local = self.server.streams()
        missing_pathnames = set([stream["pathname"]
                                for stream in remote]) - set([stream["pathname"] for stream in local])
        for stream in remote:
            if stream["pathname"] in missing_pathnames:
                self.server.make_stream(stream["pathname"], stream["name"])
                local_bundles = []
            else:
                local_bundles = [
                    bundle for bundle in self.server.bundles(stream["pathname"])]
            remote_bundles = [
                bundle for bundle in self.remote_server.bundles(stream["pathname"])]
            missing_bundles = set(
                (bundle["content_sha1"] for bundle in remote_bundles))
            missing_bundles -= set(
                (bundle["content_sha1"] for bundle in local_bundles))
            try:
                missing_bytes = sum(
                    (bundle["content_size"]
                     for bundle in remote_bundles
                     if bundle["content_sha1"] in missing_bundles))
            except KeyError as ex:
                # Older servers did not return content_size so this part is
                # optional
                missing_bytes = None
            if missing_bytes:
                print "Stream %s needs update (%s)" % (stream["pathname"], self._filesizeformat(missing_bytes))
            elif missing_bundles:
                print "Stream %s needs update (no estimate available)" % (stream["pathname"],)
            else:
                print "Stream %s is up to date" % (stream["pathname"],)
            for content_sha1 in missing_bundles:
                print "Getting %s" % (content_sha1,),
                sys.stdout.flush()
                data = self.remote_server.get(content_sha1)
                print "got %s, storing" % (self._filesizeformat(len(data["content"]))),
                sys.stdout.flush()
                try:
                    self.server.put(
                        data["content"],
                        data["content_filename"],
                        stream["pathname"])
                except xmlrpclib.Fault as ex:
                    if ex.faultCode == 409:  # duplicate
                        print "already present (in another stream)"
                    else:
                        raise
                else:
                    print "done"


class data_views(XMLRPCCommand):
    """
    Show data views defined on the server
    """
    renderer = _get_pretty_renderer(
        column_map={
            'name': 'Name',
            'summary': 'Summary',
        },
        order=('name', 'summary'),
        empty="There are no data views defined yet",
        caption="Data Views")

    def invoke_remote(self):
        self._check_server_version(self.server, "0.4")
        self.renderer.render(self.server.data_views())
        print
        print "Tip: to invoke a data view try `lc-tool query-data-view`"


class query_data_view(XMLRPCCommand):
    """
    Invoke a specified data view
    """
    @classmethod
    def register_arguments(cls, parser):
        super(query_data_view, cls).register_arguments(parser)
        parser.add_argument("QUERY", metavar="QUERY", nargs="...",
                            help="Data view name and any optional \
                            and required arguments")

    def _probe_data_views(self):
        """
        Probe the server for information about data views
        """
        with self.safety_net():
            self.use_non_legacy_api_if_possible()
            self._check_server_version(self.server, "0.4")
            return self.server.data_views()

    def reparse_arguments(self, parser, raw_args):
        self.data_views = self._probe_data_views()
        if self.data_views is None:
            return
        # Here we hack a little, the last actuin is the QUERY action added
        # in register_arguments above. By removing it we make the output
        # of lc-tool query-data-view NAME --help more consistent.
        del parser._actions[-1]
        subparsers = parser.add_subparsers(
            title="Data views available on the server")
        for data_view in self.data_views:
            data_view_parser = subparsers.add_parser(
                data_view["name"],
                help=data_view["summary"],
                epilog=data_view["documentation"])
            data_view_parser.set_defaults(data_view=data_view)
            group = data_view_parser.add_argument_group("Data view parameters")
            for argument in data_view["arguments"]:
                if argument["default"] is None:
                    group.add_argument(
                        "--{name}".format(
                            name=argument["name"].replace("_", "-")),
                        dest=argument["name"],
                        help=argument["help"],
                        type=str,
                        required=True)
                else:
                    group.add_argument(
                        "--{name}".format(
                            name=argument["name"].replace("_", "-")),
                        dest=argument["name"],
                        help=argument["help"],
                        type=str,
                        default=argument["default"])
        self.args = self.parser.parse_args(raw_args)

    def invoke(self):
        # Override and _not_ call 'use_non_legacy_api_if_possible' as we
        # already did this reparse_arguments
        with self.safety_net():
            return self.invoke_remote()

    def invoke_remote(self):
        if self.data_views is None:
            return -1
        self._check_server_version(self.server, "0.4")
        # Build a collection of arguments for data view
        data_view_args = {}
        for argument in self.args.data_view["arguments"]:
            arg_name = argument["name"]
            if arg_name in self.args:
                data_view_args[arg_name] = getattr(self.args, arg_name)
        # Invoke the data view
        response = self.server.query_data_view(
            self.args.data_view["name"],
            data_view_args)
        # Create a pretty-printer
        renderer = _get_pretty_renderer(
            caption=self.args.data_view["summary"],
            order=[item["name"] for item in response["columns"]])
        # Post-process the data so that it fits the printer
        data_for_renderer = [
            dict(zip(
                [column["name"] for column in response["columns"]],
                row))
            for row in response["rows"]]
        # Print the data
        renderer.render(data_for_renderer)


class version(Command):

    """
    Show dashboard client version
    """

    def invoke(self):
        import versiontools
        from lava_dashboard_tool import __version__
        print "Dashboard client version: {version}".format(
            version=versiontools.format_version(__version__))



# Copyright (C) 2013 Linaro Limited
#
# Author: Milo Casagrande <milo.casagrande@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""
Lava init commands.

When invoking:

 `lava init [DIR]`

the command will create a default directory and files structure as follows:

DIR/
   |
   +- JOB_FILE.json
   +- tests/
           |
           + mytest.sh
           + lavatest.yaml

If DIR is not passed, it will use the current working directory.
JOB_FILE is a file name that will be asked to the user, along with
other necessary information to define the tests.

If the user manually updates either the lavatest.yaml or mytest.sh file, it is
necessary to run the following command in order to update the job definition:

 `lava update [JOB|DIR]`
"""

import copy
import json
import os
import sys

from lava.helper.command import BaseCommand
from lava.helper.template import (
    expand_template,
    set_value
)
from lava.job import (
    JOB_FILE_EXTENSIONS,
)
from lava.job.templates import (
    LAVA_TEST_SHELL_TAR_REPO_KEY,
)
from lava.parameter import (
    Parameter,
)
from lava.testdef import (
    DEFAULT_TESTDEF_FILENAME,
)
from lava.tool.errors import CommandError
from lava_tool.utils import (
    base64_encode,
    create_dir,
    create_tar,
    edit_file,
    retrieve_file,
    write_file,
)

# Default directory structure name.
TESTS_DIR = "tests"

# Internal parameter ids.
JOBFILE_ID = "jobfile"

JOBFILE_PARAMETER = Parameter(JOBFILE_ID)
JOBFILE_PARAMETER.store = False

INIT_TEMPLATE = {
    JOBFILE_ID: JOBFILE_PARAMETER,
}


class init(BaseCommand):
    """Set-ups the base directory structure."""

    @classmethod
    def register_arguments(cls, parser):
        super(init, cls).register_arguments(parser)
        parser.add_argument("DIR",
                            help=("The name of the directory to initialize. "
                                  "Defaults to current working directory."),
                            nargs="?",
                            default=os.getcwd())

    def invoke(self):
        full_path = os.path.abspath(self.args.DIR)

        if os.path.isfile(full_path):
            raise CommandError("'{0}' already exists, and is a "
                               "file.".format(self.args.DIR))

        create_dir(full_path)
        data = self._update_data()

        # Create the directory that will contain the test definition and
        # shell script.
        test_path = create_dir(full_path, TESTS_DIR)
        shell_script = self.create_shell_script(test_path)
        # Let the user modify the file.
        edit_file(shell_script)

        testdef_file = self.create_test_definition(
            os.path.join(test_path, DEFAULT_TESTDEF_FILENAME))

        job = data[JOBFILE_ID]
        self.create_tar_repo_job(
            os.path.join(full_path, job), testdef_file, test_path)

    def _update_data(self):
        """Updates the template and ask values to the user.

        The template in this case is a layout of the directory structure as it
        would be written to disk.

        :return A dictionary containing all the necessary file names to create.
        """
        data = copy.deepcopy(INIT_TEMPLATE)
        expand_template(data, self.config)

        return data


class run(BaseCommand):
    """Runs a job on the local dispatcher."""

    @classmethod
    def register_arguments(cls, parser):
        super(run, cls).register_arguments(parser)
        parser.add_argument("JOB",
                            help=("The job file to run, or a directory "
                                  "containing a job file. If nothing is "
                                  "passed, it uses the current working "
                                  "directory."),
                            nargs="?",
                            default=os.getcwd())

    def invoke(self):
        full_path = os.path.abspath(self.args.JOB)
        job_file = retrieve_file(full_path, JOB_FILE_EXTENSIONS)

        super(run, self).run(job_file)


class submit(BaseCommand):
    """Submits a job to LAVA."""

    @classmethod
    def register_arguments(cls, parser):
        super(submit, cls).register_arguments(parser)
        parser.add_argument("JOB",
                            help=("The job file to send, or a directory "
                                  "containing a job file. If nothing is "
                                  "passed, it uses the current working "
                                  "directory."),
                            nargs="?",
                            default=os.getcwd())

    def invoke(self):
        full_path = os.path.abspath(self.args.JOB)
        job_file = retrieve_file(full_path, JOB_FILE_EXTENSIONS)

        super(submit, self).submit(job_file)


class update(BaseCommand):
    """Updates a job file with the correct data."""

    @classmethod
    def register_arguments(cls, parser):
        super(update, cls).register_arguments(parser)
        parser.add_argument("JOB",
                            help=("Automatically updates a job file "
                                  "definition. If nothing is passed, it uses"
                                  "the current working directory."),
                            nargs="?",
                            default=os.getcwd())

    def invoke(self):
        full_path = os.path.abspath(self.args.JOB)
        job_file = self.retrieve_file(full_path, JOB_FILE_EXTENSIONS)
        job_dir = os.path.dirname(job_file)
        tests_dir = os.path.join(job_dir, TESTS_DIR)

        if os.path.isdir(tests_dir):
            tar_repo = None
            try:
                tar_repo = create_tar(tests_dir)
                encoded_tests = base64_encode(tar_repo)

                json_data = None
                with open(job_file, "r") as json_file:
                    try:
                        json_data = json.load(json_file)
                        set_value(json_data, LAVA_TEST_SHELL_TAR_REPO_KEY,
                                  encoded_tests)
                    except Exception:
                        raise CommandError("Cannot read job file "
                                           "'{0}'.".format(job_file))

                content = json.dumps(json_data, indent=4)
                write_file(job_file, content)

                print >> sys.stdout, "Job definition updated."
            finally:
                if tar_repo and os.path.isfile(tar_repo):
                    os.unlink(tar_repo)
        else:
            raise CommandError("Cannot find tests directory.")



# Copyright (C) 2011 Linaro Limited
#
# Author: Michael Hudson-Doyle <michael.hudson@linaro.org>
#
# This file is part of lava-tool.
#
# lava-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-tool.  If not, see <http://www.gnu.org/licenses/>.

"""
Unit tests for the lava_tool.authtoken package
"""

import base64
import StringIO
from unittest import TestCase
import urlparse
import xmlrpclib

from mocker import ARGS, KWARGS, Mocker

from lava_tool.authtoken import (
    AuthenticatingServerProxy,
    XMLRPCTransport,
    MemoryAuthBackend,
)
from lava_tool.interface import LavaCommandError


class TestAuthenticatingServerProxy(TestCase):

    def auth_headers_for_method_call_on(self, url, auth_backend):
        parsed = urlparse.urlparse(url)

        mocker = Mocker()
        transport = mocker.mock()

        auth_data = []

        def intercept_request(host, handler, request_body, verbose=0):
            actual_transport = XMLRPCTransport(parsed.scheme, auth_backend)
            request = actual_transport.build_http_request(
                host, handler, request_body)
            if (request.has_header('Authorization')):
                auth_data.append(request.get_header('Authorization'))

        response_body = xmlrpclib.dumps((1,), methodresponse=True)
        response = StringIO.StringIO(response_body)
        response.status = 200
        response.__len__ = lambda: len(response_body)

        transport.request(ARGS, KWARGS)
        mocker.call(intercept_request)
        mocker.result(response)

        with mocker:
            server_proxy = AuthenticatingServerProxy(
                url, auth_backend=auth_backend, transport=transport)
            server_proxy.method()

        return auth_data

    def user_and_password_from_auth_data(self, auth_data):
        if len(auth_data) != 1:
            self.fail("expected exactly 1 header, got %r" % len(auth_data))
        [value] = auth_data
        if not value.startswith("Basic "):
            self.fail("non-basic auth header found in %r" % auth_data)
        auth = base64.b64decode(value[len("Basic "):])
        if ':' in auth:
            return tuple(auth.split(':', 1))
        else:
            return (auth, None)

    def test_no_user_no_auth(self):
        auth_headers = self.auth_headers_for_method_call_on(
            'http://localhost/RPC2/', MemoryAuthBackend([]))
        self.assertEqual([], auth_headers)

    def test_token_used_for_auth_http(self):
        auth_headers = self.auth_headers_for_method_call_on(
            'http://user@localhost/RPC2/',
            MemoryAuthBackend([('user', 'http://localhost/RPC2/', 'TOKEN')]))
        self.assertEqual(
            ('user', 'TOKEN'),
            self.user_and_password_from_auth_data(auth_headers))

    def test_token_used_for_auth_https(self):
        auth_headers = self.auth_headers_for_method_call_on(
            'https://user@localhost/RPC2/',
            MemoryAuthBackend([('user', 'https://localhost/RPC2/', 'TOKEN')]))
        self.assertEqual(
            ('user', 'TOKEN'),
            self.user_and_password_from_auth_data(auth_headers))

    def test_port_included(self):
        auth_headers = self.auth_headers_for_method_call_on(
            'http://user@localhost:1234/RPC2/',
            MemoryAuthBackend(
                [('user', 'http://localhost:1234/RPC2/', 'TOKEN')]))
        self.assertEqual(
            ('user', 'TOKEN'),
            self.user_and_password_from_auth_data(auth_headers))

    def test_error_when_user_but_no_token(self):
        self.assertRaises(
            LavaCommandError,
            self.auth_headers_for_method_call_on,
            'http://user@localhost/RPC2/',
            MemoryAuthBackend([]))



# Copyright (C) 2010, 2011 Linaro Limited
#
# Author: Michael Hudson-Doyle <michael.hudson@linaro.org>
#
# This file is part of lava-scheduler-tool.
#
# lava-scheduler-tool is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation
#
# lava-scheduler-tool is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lava-scheduler-tool.  If not, see <http://www.gnu.org/licenses/>.

__version__ = (0, 6, 0, "dev", 0)
